--[[
Copyright (c) 2008-2009, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local TR = LibStub( "LibLordFarlander-TableRecycler-1.0", true );

if( not TR ) then
    error( ("%s requires %s"):format( L["Calendar_Broker"], "LibLordFarlander-TableRecycler-1.0" ) );
end--if

local L = LibStub( "AceLocale-3.0" ):GetLocale( "Calendar_Broker" );

-- Constants
local CALENDARBROKER_SECPERMIN = 60;
local CALENDARBROKER_SECPERHOUR = 60 * CALENDARBROKER_SECPERMIN;
CALENDARBROKER_SECPERDAY = 24 * CALENDARBROKER_SECPERHOUR;

--Global Variables
CalendarBroker.events = {
    EventsTodayCount = 0,
};

function CalendarBroker.Time2Text( t )
    local days   = math.floor( t / CALENDARBROKER_SECPERDAY );
    local hours  = ( t - days * CALENDARBROKER_SECPERDAY ) / CALENDARBROKER_SECPERHOUR;
    local minute = ( t - ( days * CALENDARBROKER_SECPERDAY + math.floor( hours ) * CALENDARBROKER_SECPERDAY ) ) / CALENDARBROKER_SECPERMIN;
    local textT  = "ERROR";

    hours = t / CALENDARBROKER_SECPERHOUR;
    if( t <= 0 ) then
        textT = L["NOW"];
    elseif( ( days > 1 ) and ( hours > 48 ) ) then
        textT = L["%i days"]:format( days );
    elseif( hours > 1.0 ) then
        textT = L["%.1f hours"]:format( hours );
    elseif( hours == 1 ) then
        textT = L["1 hour"];
    elseif( minute > 1 ) then
        textT = L["%.1f mins"]:format( minute );
    else
        textT = L["soon"];
    end--if
    return textT;
end--CalendarBroker.Time2Text( d, h, m )

function CalendarBroker:GetFormattedDate( seconds )
    if( self.db.profile.DateFormat_US ) then
        return date( "%m/%d/%y", seconds );
    end--if
    return date( "%d/%m/%y", seconds );
end--CalendarBroker:GetFormattedDate( seconds )

function CalendarBroker:GetFormattedTime( seconds )
    if( self.db.profile.TimeFormat_12h ) then
        return date( "%I:%M %p", seconds );
    end--if
    return date( "%H:%M", seconds );
end--CalendarBroker:GetFormattedDate( seconds )

function CalendarBroker.SecondsFromEpoch( month, day, year, hour, min, sec )
    if( type( month ) == "table" ) then
        time( month );
    else
        local tim = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour or 0, "min", min or 0, "sec", sec or 0 );
        local secs = time( tim );

        TR:ReturnTable( tim );
        return secs;
    end--if
end--CalendarBroker.SecondsFromEpoch( month, day, year, hour, min, sec )

function CalendarBroker.PrintDate( seconds )
    -- Used to debug the math
    return date( "%m/%d/%y %H:%M:%S", seconds );
end--CalendarBroker.PrintDate( seconds )

function CalendarBroker:PrintDebug( text )
    if( self.db.profile.DebugMessages ) then
        self:Print( "|cffffff00DEBUG |cffffffff" .. text )
    end--if
end--CalendarBroker:PrintDebug( text )

local EVENT_TYPE_MAP = {
    [CALENDAR_EVENTTYPE_RAID or 1] = (" (%s)"):format( L["Raid"] ),
    [CALENDAR_EVENTTYPE_DUNGEON or 2] = (" (%s)"):format( L["Dungeon"] ),
    [CALENDAR_EVENTTYPE_PVP or 3] = (" (%s)"):format( L["PVP"] ),
    [CALENDAR_EVENTTYPE_MEETING or 4] = (" (%s)"):format( L["Meeting"] ),
    [CALENDAR_EVENTTYPE_OTHER or 5] = "",
};

local INVITE_TYPE_MAP = {
    [CALENDAR_INVITESTATUS_INVITED or 1] = (" (%s)"):format( L["Invited"] ),
    [CALENDAR_INVITESTATUS_ACCEPTED or 2] = (" (%s)"):format( L["Accepted"] ),
    [CALENDAR_INVITESTATUS_DECLINED or 3] = (" (%s)"):format( L["Declined"] ),
    [CALENDAR_INVITESTATUS_CONFIRMED or 4] = (" (%s)"):format( L["Confirmed"] ),
    [CALENDAR_INVITESTATUS_OUT or 5] = (" (%s)"):format( L["Out"] ),
    [CALENDAR_INVITESTATUS_STANDBY or 6] = (" (%s)"):format( L["Stand by"] ),
};

function CalendarBroker:ProcessWorldEvents()
    local currTime_Obj = date( "*t" );
    local currTime_Sec  = time( currTime_Obj );
    local ourevents = TR:GetTable();
    local counts = TR:GetTable();
    local _, month, _, year = CalendarGetDate();
    local prevmonth, prevyear, _, _ = CalendarGetMonth();

    CalendarSetAbsMonth( month, year );
    CalendarSetMonth( -1 );

    self.events.EventsTodayCount = 0;
    if( self.events.EventsToday ) then
        TR:ReturnTable( self.events.EventsToday );
    end--if
    self.events.EventsToday = TR:GetTable();
    if( self.events.EventsUpcoming ) then
        TR:ReturnTable( self.events.EventsUpcoming );
    end--if
    self.events.EventsUpcoming = TR:GetTable();

    for monthOffset = 0, 3 do
        CalendarSetMonth( 1 );

        local month, year, days, firstday = CalendarGetMonth();

        for day = 1, days do
            for i = 1, CalendarGetNumDayEvents( 0, day ) do
                local title, hour, minute, calendarType, sequenceType, eventType, texture, modStatus, inviteStatus, invitedBy, difficulty = CalendarGetDayEvent( 0, day, i );

--                    self:Print( ("CalendarGetDayEvent() = %s, %i, %i, %s, %s, %i, %s, %s, %i, %s, %i"):format( title, hour, minute, calendarType, sequenceType, eventType, texture, modStatus, inviteStatus, invitedBy, difficulty ) );
                if( ( calendarType == "RAID_RESET" ) or ( calendarType == "RAID_LOCKOUT" ) ) then
                    local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );
                    if( difficulty > 1 ) then
                        title = ("%s (%s)"):format( title, _G["DUNGEON_DIFFICULTY" .. difficulty] );
                    end--if
                    self:AddEvent( time( start ), 1 * CALENDARBROKER_SECPERHOUR, title, "RAID", "WOW" );
                    TR:ReturnTable( start );
                elseif( calendarType == "HOLIDAY" ) then
                    local event = ourevents[title .. tostring( counts[title] or 0 )];
                    local useTitle = title;
                    local location = texture:match( "Calendar_DarkmoonFaire(%w+)" );                        

                    if( location ) then
                        -- Darkmoon Faire, get the location of where it is for this time from the texture's name
                        if( location == "Elwynn" ) then
                            useTitle = ("%s (%s)"):format( title, L["Elwynn Forest"] );
                        elseif( location == "Mulgore" ) then
                            useTitle = ("%s (%s)"):format( title, L["Mulgore"] );
                        elseif( location == "Terokkar" ) then
                            useTitle = ("%s (%s)"):format( title, L["Terokkar Forest"] );
                        end--if
                    end--if                    

                    if( sequenceType == "START" ) then
                        local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );

                        ourevents[title .. tostring( counts[title] or 0 )] = TR:GetTable( "title", useTitle, "start", start , "length", ( 23 - hour ) * CALENDARBROKER_SECPERHOUR + ( 60 - minute ) * CALENDARBROKER_SECPERMIN, "type", calendarType );
                    elseif( sequenceType == "END" ) then
                        if( event ) then
                            event.length = event.length + hour * CALENDARBROKER_SECPERHOUR + minute * CALENDARBROKER_SECPERMIN;
                        else
                            local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );
    
                            ourevents[title .. tostring( counts[title] or 0 )] = TR:GetTable( "title", useTitle, "start", start , "length", ( 23 - hour ) * CALENDARBROKER_SECPERHOUR + ( 60 - minute ) * CALENDARBROKER_SECPERMIN, "type", calendarType );
                        end--if
                        counts[title] = ( counts[title] or 0 ) + 1;
                    else
                        -- Not sure what to do about "INFO"
                        if( event ) then
                            event.length = event.length + 24 * CALENDARBROKER_SECPERHOUR;
                        else
                            local length = ( 23 - hour ) * CALENDARBROKER_SECPERHOUR + ( 60 - minute ) * CALENDARBROKER_SECPERMIN;
                            local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );
    
                            if( ( hour == 14 ) and ( minute == 0 ) ) then
                                length = 2 * CALENDARBROKER_SECPERHOUR;
                            end--if
                            ourevents[title .. tostring( counts[title] or 0 )] = TR:GetTable( "title", useTitle, "start", start , "length", length, "type", calendarType );
                            if( sequenceType ~= "ONGOING" ) then
                                counts[title] = ( counts[title] or 0 ) + 1;
                            end--if
                        end--if
                    end--if
                elseif( ( calendarType == "PLAYER" ) and ( inviteStatus ~= CALENDAR_INVITESTATUS_DECLINED ) and ( inviteStatus ~= CALENDAR_INVITESTATUS_OUT ) ) then
                    local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );

                    self:AddEvent( time( start ), 1 * CALENDARBROKER_SECPERHOUR, title, calendarType, "WOW" );
                    TR:ReturnTable( start );
                elseif( ( calendarType == "GUILD" ) and ( inviteStatus ~= CALENDAR_INVITESTATUS_DECLINED ) and ( inviteStatus ~= CALENDAR_INVITESTATUS_OUT ) ) then
                    local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );

                    self:AddEvent( time( start ), 1 * CALENDARBROKER_SECPERHOUR, title, calendarType, "WOW" );
                    TR:ReturnTable( start );
                elseif( calendarType == "SYSTEM" ) then
                    local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );

                    self:AddEvent( time( start ), 1 * CALENDARBROKER_SECPERHOUR, title, calendarType, "WOW" );
                    TR:ReturnTable( start );
                elseif( calendarType == "ARENA" ) then
                    local start = TR:GetTable( "year", year, "month", month, "day", day, "hour", hour, "min", minute, "sec", 0 );

                    self:AddEvent( time( start ), 1 * CALENDARBROKER_SECPERHOUR, title, calendarType, "WOW" );
                    TR:ReturnTable( start );
                end--if
            end--for
        end--for
    end--for

    CalendarSetAbsMonth( prevmonth, prevyear );
    for _, event in pairs( ourevents ) do
        self:AddEvent( time( event.start ), event.length, event.title, event.type, "WOW" );
    end--for
    TR:ReturnTable( ourevents );
    TR:ReturnTable( counts );

    -- Since different OS and different timezones may have different points of Epoch,
    -- everything is referenced to the release of WOW (Azeroth's epoch)
    self.db.profile.ZuluOffset = self.SecondsFromEpoch( 11, 23, 2004 );

    -- If in battleground may need to offset servertime
    local ServerHour, ServerMin = GetGameTime();
    local BGflag = false;

    for i = 1, MAX_BATTLEFIELD_QUEUES do
        local status, _, _, _, _, _ = GetBattlefieldStatus( i );

        BGflag = ( status == "active" ) or BGflag;
    end--for
    if( BGflag ) then
        ServerHour = currTime_Obj.hour - self.db.profile.ServerTimeOffset;
    else
        self.db.profile.ServerTimeOffset = currTime_Obj.hour - ServerHour;
    end--if

    -- if GEM is loaded, then add the events that I am subscribed to
    if( GEM3_Events == nil ) then
        self.db.profile.UseGEMdatabase = false;
    end--if

    if( self.db.profile.UseGEMdatabase and GEM3_Events[GEM3_Realm] ) then
        for i, event in ipairs( GEM3_Events[GEM3_Realm] ) do
            local EVloc = event.ev_place;
            local EVsec = event.ev_date + self.db.profile.ServerTimeOffset * CALENDARBROKER_SECPERHOUR;
            local ET = date( "*t", EVsec );
            local EVtype = event.ev_type;
            local type = "GEM";

            if( EVtype == GEM3_EVT_TYPE_NONE ) then
                type = "OTHER";
            elseif( EVtype == GEM3_EVT_TYPE_TEST ) then
                type = "OTHER";
            elseif( EVtype == GEM3_EVT_TYPE_RESET ) then
                type = "RAID";
            elseif( EVtype == GEM3_EVT_TYPE_LFG ) then
                type = "PLAYER";
            elseif( EVtype == GEM3_EVT_TYPE_POLL ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_PLAYER_AVALABILITY ) then
                type = "PLAYER";
            elseif( EVtype == GEM3_EVT_TYPE_RAID_AVAILABILITY ) then
                type = "RAID";
            elseif( EVtype == GEM3_EVT_TYPE_INSTANCE5 ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_INSTANCE5_BC ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_INSTANCE5_WOTLK ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_INSTANCE5_BCH ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_INSTANCE5_WOTLKH ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_RAID ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_RAID_BC ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_RAID_WOTLK ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_PVP ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_GUILD_EVENT ) then
                type = "GUILD";
            elseif( EVtype == GEM3_EVT_TYPE_GUILD_OFFICERMEETING ) then
                type = "GUILD";
            end--if
            self:AddEvent( EVsec, 1 * CALENDARBROKER_SECPERHOUR, EVloc, type, "GEM" );
        end--for
    end--if

    -- Sort the data
    table.sort( self.events.EventsToday, function( a, b ) return a.time2 < b.time2 end );
    table.sort( self.events.EventsUpcoming, function( a, b ) return a.time1 < b.time1 end );
end--CalendarBroker:ProcessWorldEvents()

function CalendarBroker:AddEvent( start, duration, text, type, calendar )
    -- start and duration are in units of seconds
    local current = time();
    local EndTime = start + duration;

    if( not type ) then
        type = "NONE";
        calendar = "";
    end--if
--    self:PrintDebug( "Adding event " .. text .. " " .. date( "%m/%d/%y %H:%M:%S", start ) .. " to " .. date( "%m/%d/%y %H:%M:%S", EndTime ) );

    if( current < EndTime ) then
        -- Has not ended
        if( current >= start ) then
            table.insert( self.events.EventsToday, TR:GetTable( "time1", 0, "title", text, "type", type, "time2", EndTime, "source", calendar, "length", duration ) );
            self.events.EventsTodayCount = self.events.EventsTodayCount + 1;
--            self:PrintDebug( "Current event" );
        else
            table.insert( self.events.EventsUpcoming, TR:GetTable( "time1", start - current, "title", text, "type", type, "time2", start, "source", calendar, "length", duration ) );
--            self:PrintDebug( "Future event " );
        end--if
--    else
--        self:PrintDebug( "Already over" );
    end--if
end--CalendarBroker:AddEvent( start, duration, text )
